
from hyperspy.misc.example_signals_loading import load_1D_EDS_SEM_spectrum as\
    EDS_SEM_Spectrum

from hyperspy.misc.example_signals_loading import load_1D_EDS_TEM_spectrum as\
    EDS_TEM_Spectrum
