# -*- coding: utf-8 -*-
"""Common docstring snippets.

"""

START_HSPY = \
    """When starting HyperSpy using the ``hyperspy`` script (e.g. by executing
``hyperspy`` in a console, using the context menu entries or using the links in
the ``Start Menu``, the :mod:`~hyperspy.api` package is imported in the user
namespace as ``hs``, i.e. by executing the following:

    >>> import hyperspy.api as hs


(Note that code snippets are indicated by three greater-than signs)

We recommend to import the HyperSpy API as above also when doing it manually.
The docstring examples assume that `hyperspy` has been imported as `hs`,
numpy as ``np`` and ``matplotlib.pyplot`` as ``plt``. """
