class VisibleDeprecationWarning(UserWarning):

    """Visible deprecation warning.
    By default, python will not show deprecation warnings, so this class
    provides a visible one.

    """
    pass
