# Global constants
# Fundamental constants


R = 13.6056923  # Rydberg of energy in eV
e = 1.602176487E-19  # electron charge in C
m0 = 9.10938215E-31  # electron rest mass in kg
a0 = 5.2917720859E-11  # Bohr radius in m
c = 2997.92458E8  # speed of light in m/s
