"""Datasets.

Submodules:


The :mod:`~hyperspy.api.datasets` module contains the following submodules:

    :mod:`~hyperspy.api.datasets.example_signals`
        Example datasets distributed with HyperSpy.



"""
